left_boundary_values = "Left boundary values"
right_boundary_values = "Right boundary values"
initial_values = "Initial values"
initial_derivatives = "Initial Derivatives"
source_term = "Source term (optional)"
wave_speed = "Wave speed"
length = "Length"
time_interval = "Time interval"
samples = "Samples"

left_boundary_values_symbol = "φ₁(t)"
right_boundary_values_symbol = "φ₂(t)"
initial_values_symbol = "u(0, x)"
initial_derivatives_symbol = "∂u(0, x)/∂t"
source_term_symbol = "s(t, x)"
wave_speed_symbol = "c"
length_symbol = "L"
time_interval_symbol = "T"
samples_symbol = "N"
